import asyncio
import json
import os
import traceback
from abc import ABC, abstractmethod
from contextlib import AsyncExitStack
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from random import shuffle

from pptagent.agent import Agent
from pptagent.apis import API_TYPES, CodeExecutor
from pptagent.document import Document
from pptagent.llms import AsyncLLM
from pptagent.presentation import (
    GroupShape,
    Layout,
    Picture,
    Presentation,
    SlidePage,
    StyleArg,
)
from pptagent.response import EditorOutput, LayoutChoice, Outline, OutlineItem
from pptagent.utils import (
    Language,
    edit_distance,
    get_logger,
    tenacity_decorator,
)

logger = get_logger(__name__)

style = StyleArg.all_true()
style.area = False


def get_length_factor(src_lan: Language, dst_lang: Language):
    if src_lan.latin == dst_lang.latin:  # belong to the same language family
        return 1.2
    elif src_lan.latin:  # source is latin, dst is cjk
        return 0.7
    else:  # source is cjk, dst is latin
        return 2.0


class FunctionalLayouts(Enum):
    OPENING = "opening"
    TOC = "table of contents"
    SECTION_OUTLINE = "section outline"
    ENDING = "ending"


FunctionalContent = {
    FunctionalLayouts.OPENING.value: "This slide is a presentation opening, presenting available meta information, like title, author, date, etc.",
    FunctionalLayouts.TOC.value: "This slide is the Table of Contents, outlining the presentation's sections. Please use the Table of Contents given in the retrieved content, remove numbering and ensure the completeness of the Table of Contents, and generate the final output with the language specified in the input.",
    FunctionalLayouts.SECTION_OUTLINE.value: "This slide marks the beginning of a new section and should present the content from <title>{}</title> as the section title clearly, any existing prefix or numbering should be removed. If section number is provided in the schema, use <section_number>{}</section_number>. And provide a brief summary of the section if additional elements are provided in the schema.",
    FunctionalLayouts.ENDING.value: "This slide is an *ending slide*, simply express your gratitude like 'Thank you!' or '谢谢' as the main title and *do not* include other meta information if not specified.",
}


@dataclass
class PPTGen(ABC):
    """
    Stage II: Presentation Generation
    An base class for generating PowerPoint presentations.
    It accepts a reference presentation as input, then generates a presentation outline and slides.
    """

    roles = [
        "editor",
        "coder",
        "content_organizer",
        "layout_selector",
    ]

    language_model: AsyncLLM
    vision_model: AsyncLLM
    retry_times: int = 3
    sim_bound: float = 0.5
    force_pages: bool = False
    error_exit: bool = False
    record_cost: bool = False
    _initialized: bool = False

    def __post_init__(self):
        self._hire_staffs(self.record_cost, self.language_model, self.vision_model)

    def set_reference(
        self,
        slide_induction: dict,
        presentation: Presentation,
        hide_small_pic_ratio: float | None = 0.2,
        keep_in_background: bool = True,
    ):
        """
        Set the reference presentation and extracted presentation information.

        Args:
            presentation (Presentation): The presentation object.
            slide_induction (dict): The slide induction data.

        Returns:
            PPTGen: The updated PPTGen object.
        """
        self.presentation = presentation

        self.reference_lang = Language(**slide_induction.pop("language"))
        self.functional_layouts = slide_induction.pop("functional_keys")

        self.layouts: dict[str, Layout] = {
            k: Layout(title=k, **v) for k, v in slide_induction.items()
        }
        self.empty_prs = deepcopy(self.presentation)
        assert hide_small_pic_ratio is None or hide_small_pic_ratio > 0, (
            "hide_small_pic_ratio must be positive or None"
        )
        if hide_small_pic_ratio is not None:
            self._hide_small_pics(hide_small_pic_ratio, keep_in_background)

        self.text_layouts = [
            k
            for k in self.layouts.keys()
            if k.endswith("text") and k not in self.functional_layouts
        ]
        self.multimodal_layouts = [
            k
            for k in self.layouts.keys()
            if not k.endswith("text") and k not in self.functional_layouts
        ]
        if len(self.text_layouts) == 0:
            self.text_layouts = self.multimodal_layouts
        if len(self.multimodal_layouts) == 0:
            self.multimodal_layouts = self.text_layouts

        self._initialized = True
        return self

    async def generate_pres(
        self,
        source_doc: Document,
        num_slides: int | None = None,
        outline: list[OutlineItem] | None = None,
        image_dir: str | None = None,
        dst_language: Language | None = None,
        length_factor: float | None = None,
        auto_length_factor: bool = True,
        max_at_once: int | None = None,
    ):
        """
        Generate a PowerPoint presentation.

        Args:
            source_doc (Document): The source document.
            num_slides (int | None): The number of slides to generate.
            outline (list[OutlineItem] | None): The outline of the presentation.
            image_dir (str | None): The directory of the images.
            dst_language (Language | None): The destination language.
            length_factor (float | None): The length factor.
            auto_length_factor (bool): Whether to automatically calculate the length factor.
            max_at_once (int | None): The maximum number of slides to generate at once.

        Returns:
            tuple[Presentation, dict]: A tuple containing the generated presentation and the history of the agents.
        """
        # validate image existence
        source_doc.validate_medias(image_dir)
        source_doc.metadata["presentation-date"] = datetime.now().strftime("%Y-%m-%d")
        assert self._initialized, "PPTAgent not initialized, call `set_reference` first"
        self.source_doc = source_doc
        length_factor = length_factor or os.getenv("PPTAGENT_LENGTH_FACTOR", None)
        if (
            auto_length_factor or os.getenv("PPTAGENT_AUTO_LENGTH_FACTOR", False)
        ) and length_factor is None:
            self.dst_lang = dst_language or source_doc.language
            self.length_factor = get_length_factor(self.reference_lang, self.dst_lang)
        else:
            self.length_factor = length_factor
        succ_flag = True
        if outline is None:
            self.outline = await self.generate_outline(num_slides, source_doc)
        else:
            self.outline = outline
        pre_section = None
        section_idx = 0
        self.simple_outline = ""
        for slide_idx, item in enumerate(self.outline):
            if item.topic != pre_section and item.topic != "Functional":
                section_idx += 1
                self.simple_outline += f"Section {section_idx}: {item.topic}\n"
                pre_section = item.topic
            if item.purpose == FunctionalLayouts.SECTION_OUTLINE.value:
                item.indexes.append(section_idx)
            self.simple_outline += f"Slide {slide_idx + 1}: {item.purpose}\n"
        logger.debug(f"==========Outline Generated==========\n{self.simple_outline}")

        if max_at_once:
            semaphore = asyncio.Semaphore(max_at_once)
        else:
            semaphore = AsyncExitStack()

        slide_tasks = []
        for slide_idx, outline_item in enumerate(self.outline):
            if self.force_pages and slide_idx == num_slides:
                break
            slide_tasks.append(
                self.generate_slide(slide_idx, outline_item, semaphore=semaphore)
            )

        slide_results = await asyncio.gather(*slide_tasks, return_exceptions=True)

        generated_slides = []
        code_executors = []
        for result in slide_results:
            if isinstance(result, Exception):
                if self.error_exit:
                    succ_flag = False
                    break
                continue
            if result is not None:
                slide, code_executor = result
                generated_slides.append(slide)
                code_executors.append(code_executor)

        history = self._collect_history(
            sum(code_executors, start=CodeExecutor(self.retry_times))
        )

        if succ_flag:
            self.empty_prs.slides = generated_slides
            prs = self.empty_prs
        else:
            prs = None

        self.empty_prs = deepcopy(self.presentation)
        return prs, history

    async def generate_outline(
        self,
        num_slides: int,
        source_doc: Document,
    ):
        """
        Asynchronously generate an outline for the presentation.
        """
        assert self._initialized, (
            "AsyncPPTAgent not initialized, call `set_reference` first"
        )
        _, outline = await self.staffs["planner"](
            num_slides=num_slides,
            document_overview=source_doc.get_overview(),
            response_format=Outline.response_model(source_doc),
        )
        outline = [OutlineItem(**o) for o in outline["outline"]]
        return self._add_functional_layouts(outline)

    @abstractmethod
    def generate_slide(
        self, slide_idx: int, outline_item: OutlineItem, semaphore: AsyncExitStack
    ) -> tuple[SlidePage, CodeExecutor]:
        """
        Generate a slide from the outline item.
        """
        raise NotImplementedError("Subclass must implement this method")

    def _add_functional_layouts(self, outline: list[OutlineItem]):
        """
        Add functional layouts to the outline.
        """
        toc = []
        for item in outline:
            if item.topic not in toc and item.topic != "Functional":
                toc.append(item.topic)
        self.toc = "\n".join(toc)

        fixed_functional_slides = [
            (FunctionalLayouts.TOC.value, 0),  # toc should be inserted before opening
            (FunctionalLayouts.OPENING.value, 0),
            (FunctionalLayouts.ENDING.value, 999999),  # append to the end
        ]
        for title, pos in fixed_functional_slides:
            layout = max(
                self.functional_layouts,
                key=lambda x: edit_distance(x.lower(), title),
            )
            if edit_distance(layout, title) > 0.7:
                outline.insert(
                    pos,
                    OutlineItem(
                        purpose=layout, topic="Functional", indexes=[], images=[]
                    ),
                )

        section_layout = FunctionalLayouts.SECTION_OUTLINE.value
        section_outline = max(
            self.functional_layouts,
            key=lambda x: edit_distance(x, section_layout),
        )
        if not edit_distance(section_outline, section_layout) > 0.7:
            return outline
        full_outline = []
        pre_section = None
        for item in outline:
            if item.topic == "Functional":
                full_outline.append(item)
                continue
            if item.topic != pre_section:
                new_item = OutlineItem(
                    purpose=section_outline,
                    topic="Functional",
                    indexes=[item.topic],
                    images=[],
                )
                full_outline.append(new_item)
            full_outline.append(item)
            pre_section = item.topic
        return full_outline

    def _hide_small_pics(self, area_ratio: float, keep_in_background: bool):
        for layout in list(self.layouts.values()):
            template_slide = self.presentation.slides[layout.template_id - 1]
            pictures: list[tuple[SlidePage | GroupShape, Picture]] = list(
                template_slide.shape_filter(Picture, return_father=True)
            )
            if len(pictures) == 0:
                continue
            for father, pic in pictures:
                if pic.area / pic.slide_area < area_ratio:
                    father.shapes.remove(pic)
                    if keep_in_background:
                        template_slide.backgrounds.append(pic)
                    layout.remove_item(pic.caption)

            if len(list(template_slide.shape_filter(Picture))) == 0:
                logger.debug(
                    "All pictures in layout %s are too small, set to pure text layout",
                    layout.title,
                )
                self.layouts[layout.title.replace(":image", ":text")] = (
                    self.layouts.pop(layout.title)
                )

    def _collect_history(self, code_executor: CodeExecutor):
        """
        Collect the history of code execution, API calls and agent steps.

        Returns:
            dict: The collected history data.
        """
        history = {
            "agents": {},
            "command_history": code_executor.command_history,
            "code_history": code_executor.code_history,
            "api_history": code_executor.api_history,
        }

        for role_name, role in self.staffs.items():
            history["agents"][role_name] = role.history
            role._history = []

        return history

    def _hire_staffs(
        self,
        record_cost: bool,
        language_model: AsyncLLM,
        vision_model: AsyncLLM,
    ) -> dict[str, Agent]:
        """
        Initialize agent roles and their models
        """
        llm_mapping = {
            "language": language_model,
            "vision": vision_model,
        }
        self.staffs = {
            role: Agent(
                role,
                record_cost=record_cost,
                llm_mapping=llm_mapping,
            )
            for role in ["planner"] + self.roles
        }


class PPTAgent(PPTGen):
    """
    Asynchronous subclass of PPTGen that uses Agent for concurrent processing.
    """

    async def generate_slide(
        self, slide_idx: int, outline_item: OutlineItem, semaphore: AsyncExitStack
    ) -> tuple[SlidePage, CodeExecutor]:
        """
        Asynchronously generate a slide from the outline item.
        """
        async with semaphore:
            if outline_item.topic == "Functional":
                layout = self.layouts[outline_item.purpose]
                slide_desc = FunctionalContent[outline_item.purpose]
                if outline_item.purpose == FunctionalLayouts.SECTION_OUTLINE.value:
                    section, sec_idx = outline_item.indexes
                    slide_desc = slide_desc.format(section, sec_idx + 1)
                    outline_item.purpose = f"Section Outline of {section}"
                    outline_item.indexes = []
                    slide_content = (
                        "Document Structure:\n"
                        + self.source_doc.get_overview(
                            include_summary=True, include_image=False
                        )
                    )
                elif outline_item.purpose == FunctionalLayouts.TOC.value:
                    slide_content = "Table of Contents:\n" + self.toc
                else:
                    slide_content = "This slide is a functional layout, please follow the slide description and content schema to generate the slide content."
                header, _, _ = outline_item.retrieve(slide_idx, self.source_doc)
                header += slide_desc
            else:
                layout, header, slide_content = await self._select_layout(
                    slide_idx, outline_item
                )
            try:
                command_list, template_id = await self._generate_content(
                    layout, slide_content, header
                )
                slide, code_executor = await self._edit_slide(command_list, template_id)
            except Exception as e:
                logger.error(f"Failed to generate slide {slide_idx}, error: {e}")
                traceback.print_exc()
                raise e
            return slide, code_executor

    @tenacity_decorator
    async def _select_layout(
        self, slide_idx: int, outline_item: OutlineItem
    ) -> tuple[Layout, str, str]:
        """
        Asynchronously select a layout for the slide.
        """
        header, content_source, images = outline_item.retrieve(
            slide_idx, self.source_doc
        )
        if len(content_source) == 0:
            key_points = []
        else:
            _, key_points = await self.staffs["content_organizer"](
                content_source=content_source
            )
        slide_content = json.dumps(key_points, indent=2, ensure_ascii=False)
        layouts = self.text_layouts
        if len(images) > 0:
            slide_content += "\nImages:\n" + "\n".join(images)
            layouts = self.multimodal_layouts

        shuffle(layouts)
        _, layout_selection = await self.staffs["layout_selector"](
            outline=self.simple_outline,
            slide_description=header,
            slide_content=slide_content,
            available_layouts=layouts,
            response_format=LayoutChoice.response_model(layouts),
        )
        layout = layout_selection["layout"]
        if "image" not in layout and len(images) > 0:
            slide_content = slide_content[: slide_content.rfind("\nImages:\n")]
        return self.layouts[layout], header, slide_content

    async def _generate_content(
        self,
        layout: Layout,
        slide_content: str,
        slide_description: str,
    ) -> tuple[list, int]:
        """
        Asynchronously generate content for the slide.
        """
        elements = [el.name for el in layout.elements]
        turn_id, editor_output = await self.staffs["editor"](
            outline=self.simple_outline,
            slide_description=slide_description,
            metadata=self.source_doc.metainfo,
            slide_content=slide_content,
            schema=layout.content_schema,
            language=self.dst_lang.lid,
            response_format=EditorOutput.response_model(elements),
        )
        editor_output = EditorOutput(**editor_output)
        await self._validate_content(editor_output, layout, turn_id)
        command_list, template_id = self._generate_commands(editor_output, layout)
        return command_list, template_id

    async def _edit_slide(
        self, command_list: list, template_id: int
    ) -> tuple[SlidePage, CodeExecutor]:
        """
        Asynchronously edit the slide.
        """
        code_executor = CodeExecutor(self.retry_times)
        code_executor.command_history.append(command_list)
        turn_id, edit_actions = await self.staffs["coder"](
            api_docs=code_executor.get_apis_docs(API_TYPES.Agent.value),
            edit_target=self.presentation.slides[template_id - 1].to_html(),
            command_list="\n".join([str(i) for i in command_list]),
        )

        for error_idx in range(self.retry_times):
            edit_slide: SlidePage = deepcopy(self.presentation.slides[template_id - 1])
            feedback = code_executor.execute_actions(
                edit_actions, edit_slide, self.source_doc
            )
            if feedback is None:
                break
            logger.warning(
                "Failed to generate slide, tried %d/%d times, error: %s\n%s",
                error_idx + 1,
                self.retry_times,
                str(feedback[1]),
                edit_actions,
            )
            if error_idx == self.retry_times:
                raise Exception(
                    f"Failed to generate slide, tried too many times at editing\ntraceback: {feedback[1]}"
                )
            edit_actions = await self.staffs["coder"].retry(
                feedback[0], feedback[1], turn_id, error_idx + 1
            )
        self.empty_prs.validate(edit_slide)
        return edit_slide, code_executor

    async def _validate_content(
        self, editor_output: EditorOutput, layout: Layout, turn_id: int, retry: int = 0
    ):
        """
        Asynchronously generate commands for editing the slide content.

        Args:
            editor_output (dict): The editor output.
            layout (Layout): The layout object containing content schema.
            turn_id (int): The turn ID for retrying.
            retry (int, optional): The number of retries. Defaults to 0.

        Returns:
            list: A list of commands.

        Raises:
            Exception: If command generation fails.
        """
        try:
            allowed_images = [m.path for m in self.source_doc.iter_medias()]
            layout.validate(editor_output, allowed_images)
            if self.length_factor is not None:
                await layout.length_rewrite(
                    editor_output, self.length_factor, self.language_model
                )
        except Exception as e:
            if retry < self.retry_times:
                new_output = await self.staffs["editor"].retry(
                    e,
                    traceback.format_exc(),
                    turn_id,
                    retry + 1,
                    response_format=EditorOutput,
                )
                return await self._validate_content(
                    EditorOutput(**new_output), layout, turn_id, retry + 1
                )
            else:
                raise Exception(
                    f"Failed to generate commands, tried too many times at editing\ntraceback: {e}"
                )

    def _generate_commands(self, editor_output: EditorOutput, layout: Layout):
        command_list = []
        template_id, old_data = layout.index_template_slide(editor_output)
        for el_name, old_content in old_data.items():
            new_content = (
                editor_output[el_name].data if el_name in editor_output else []
            )
            quantity_change = len(new_content) - len(old_content)
            command_list.append(
                (
                    el_name,
                    layout[el_name].type,
                    f"quantity_change: {quantity_change}",
                    old_content,
                    new_content,
                )
            )
        return command_list, template_id



"""PPTAgent: Generating and Evaluating Presentations Beyond Text-to-Slides.

This package provides tools to automatically generate presentations from documents,
following a two-phase approach of Analysis and Generation.

For more information, visit: https://github.com/icip-cas/PPTAgent
"""

__version__ = "0.2.0"
__author__ = "Hao Zheng"
__email__ = "wszh712811@gmail.com"


# Check the version of python and python-pptx

from packaging.version import Version
from pptx import __version__ as PPTXVersion

try:
    PPTXVersion, Mark = PPTXVersion.split("+")
    assert Version(PPTXVersion) >= Version("1.0.4") and Mark == "PPTAgent"
except Exception as _:
    raise ImportError(
        "You should install the customized `python-pptx` for this project, see https://github.com/Force1ess/python-pptx"
    )

# __init__.py
from .pptgen import PPTAgent
from .document import Document
from .presentation import Presentation
from .utils import Config, Language
from .model_utils import ModelManager
from .multimodal import ImageLabler
from .llms import LLM, AsyncLLM
from .mcp_server import PPTAgentServer

__all__ = [
    "__version__",
    "__author__",
    "__email__",
    "PPTAgent",
    "PPTAgentServer",
    "Document",
    "Presentation",
    "Config",
    "Language",
    "ModelManager",
    "ImageLabler",
    "LLM",
    "AsyncLLM",
]



from test.conftest import test_config

from pptagent.multimodal import ImageLabler
from pptagent.presentation import Presentation
from os.path import join


def test_load_captions():
    prs = Presentation.from_file(
        join(test_config.template, "source.pptx"), test_config.config
    )
    image_labler = ImageLabler(prs, test_config.config)
    image_labler.apply_stats(test_config.get_image_stats())



from test.conftest import test_config

import pytest

from pptagent.document import Document


@pytest.mark.asyncio
@pytest.mark.llm
async def test_document():
    with open(f"{test_config.document}/source.md", encoding="utf-8") as f:
        markdown_content = f.read()
    cutoff = markdown_content.find("## When (and when not) to use agents")
    image_dir = test_config.document
    await Document.from_markdown(
        markdown_content[:cutoff],
        test_config.language_model,
        test_config.vision_model,
        image_dir,
    )


def test_document_from_dict():
    document = Document(**test_config.get_document_json())
    document.get_overview(include_summary=True)
    document.metainfo



from contextvars import ContextVar
from typing import Literal

from pydantic import BaseModel, Field, create_model

from pptagent.document import Document, SubSection
from pptagent.utils import get_logger

_empty_images = ContextVar(
    "_empty_images",
)

logger = get_logger(__name__)


class DocumentIndex(BaseModel):
    section: str
    subsections: list[str]

    @classmethod
    def response_model(cls, section_fields: list[str], subsection_fields: list[str]):
        return create_model(
            cls.__name__,
            section=(Literal[*section_fields], Field(...)),  # type: ignore
            subsections=(list[Literal[*subsection_fields]], Field(...)),  # type: ignore
            __base__=BaseModel,
        )


class OutlineItem(BaseModel):
    purpose: str
    topic: str
    indexes: list[DocumentIndex] | list[str]
    images: list[str] = Field(default_factory=list)

    def model_post_init(self, _) -> None:
        if self.images and not _empty_images.get():
            self.images.clear()

    def retrieve(self, slide_idx: int, document: Document):
        subsections = []
        for index in self.indexes:
            for subsection in index.subsections:
                subsections.append(document[index.section][subsection])
        header = f"Current Slide: {self.purpose}\n"
        header += f"This is the {slide_idx + 1} slide of the presentation.\n"
        content = ""
        for subsection in subsections:
            content += f"Paragraph: {subsection.title}\nContent: {subsection.content}\n"
        images = [
            f"<image>{path}</image>: {document.find_media(path=path).caption}"
            for path in self.images
        ]
        return header, content, images

    @classmethod
    def response_model(cls, document: Document):
        sections = []
        subsections = []
        for sec in document.sections:
            sections.append(sec.title)
            for subsec in sec.content:
                if isinstance(subsec, SubSection):
                    subsections.append(subsec.title)
        allowed_images = [m.path for m in document.iter_medias()]
        if allowed_images:
            literal_image = Literal[*allowed_images]  # type: ignore
            _empty_images.set(True)
        else:
            _empty_images.set(False)
            literal_image = str

        return create_model(
            cls.__name__,
            purpose=(str, Field(...)),
            topic=(str, Field(...)),
            indexes=(
                list[DocumentIndex.response_model(sections, subsections)],
                Field(...),
            ),
            images=(list[literal_image], Field(default_factory=list)),
            __base__=BaseModel,
        )


class Outline(BaseModel):
    outline: list[OutlineItem]

    @classmethod
    def response_model(cls, document: Document):
        return create_model(
            cls.__name__,
            outline=(
                list[OutlineItem.response_model(document)],
                Field(...),
            ),
            __base__=BaseModel,
        )



from .layout import Element, Layout
from .presentation import Presentation, SlidePage
from .shapes import (
    SHAPECAST,
    Background,
    Closure,
    ClosureType,
    Fill,
    Font,
    FreeShape,
    GroupShape,
    Line,
    Paragraph,
    Picture,
    Placeholder,
    SemanticPicture,
    ShapeElement,
    StyleArg,
    TextBox,
    TextFrame,
    UnsupportedShape,
)

__all__ = [
    "Presentation",
    "SlidePage",
    "SHAPECAST",
    "Background",
    "Closure",
    "ClosureType",
    "Element",
    "Fill",
    "Font",
    "FreeShape",
    "GroupShape",
    "Layout",
    "Line",
    "Paragraph",
    "Picture",
    "Placeholder",
    "SemanticPicture",
    "ShapeElement",
    "StyleArg",
    "TextBox",
    "TextFrame",
    "UnsupportedShape",
]



from dataclasses import asdict, dataclass
from functools import partial
from math import ceil

import yaml
from jinja2 import Environment, StrictUndefined, Template
from PIL import Image
from pydantic import BaseModel
from torch import Tensor

from pptagent.llms import AsyncLLM, ThinkMode
from pptagent.utils import get_json_from_response, package_join

RETRY_TEMPLATE = Template(
    """The previous output is invalid, please carefully analyze the traceback and feedback information, correct errors happened before.
            feedback:
            {{feedback}}
            traceback:
            {{traceback}}
            Give your corrected output in the same format without including the previous output:
            """
)


@dataclass
class Turn:
    """
    A class to represent a turn in a conversation.
    """

    id: int
    prompt: str
    response: str
    message: list
    retry: int = -1
    images: list[str] = None
    input_chars: int = 0
    output_chars: int = 0
    embedding: Tensor = None

    def to_dict(self):
        return {k: v for k, v in asdict(self).items() if k != "embedding"}

    def calc_token(self):
        """
        Calculate the number of tokens for the turn.
        """
        if self.images is not None:
            self.input_chars += calc_image_tokens(self.images)
        self.input_chars += len(self.prompt)
        self.output_chars = len(self.response)

    def __eq__(self, other):
        return self is other


class Agent:
    """
    An agent, defined by its instruction template and model.
    """

    def __init__(
        self,
        name: str,
        llm_mapping: dict[str, AsyncLLM],
        record_cost: bool = False,
        config: dict | None = None,
        env: Environment | None = None,
    ):
        """
        Initialize the Agent.

        Args:
            name (str): The name of the role.
        llm_mapping (dict): The mapping of the language model.
            record_cost (bool): Whether to record the token cost.
            llm (LLM): The language model.
            config (dict): The configuration.
            env (Environment): The Jinja2 environment.
        """
        self.name = name
        self.config = config
        if self.config is None:
            with open(package_join("roles", f"{name}.yaml"), encoding="utf-8") as f:
                self.config = yaml.safe_load(f)
                assert isinstance(self.config, dict), "Agent config must be a dict"
        self.llm_mapping = llm_mapping
        self.llm = self.llm_mapping[self.config["use_model"]]
        self.model = self.llm.model
        self.record_cost = record_cost
        self.return_json = self.config.get("return_json", False)
        self.system_message = self.config["system_prompt"]
        self.prompt_args = set(self.config["jinja_args"])
        self.env = env
        if self.env is None:
            self.env = Environment(undefined=StrictUndefined)
        self.template = self.env.from_string(self.config["template"])
        self.input_tokens = 0
        self.output_tokens = 0
        self._history: list[Turn] = []
        run_args = self.config.get("run_args", {})
        self.llm.__call__ = partial(self.llm.__call__, **run_args)
        self.system_tokens = len(self.system_message)

    def calc_cost(self, turns: list[Turn]):
        """
        Calculate the cost of a list of turns.
        """
        for turn in turns[:-1]:
            self.input_tokens += turn.input_chars
            self.input_tokens += turn.output_chars
        self.input_tokens += turns[-1].input_chars
        self.output_tokens += turns[-1].output_chars
        self.input_tokens += self.system_tokens

    @property
    def next_turn_id(self):
        if len(self._history) == 0:
            return 0
        return max(t.id for t in self._history) + 1

    @property
    def history(self):
        return sorted(self._history, key=lambda x: (x.id, x.retry))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name}, model={self.model})"

    async def retry(
        self,
        feedback: str,
        traceback: str,
        turn_id: int,
        error_idx: int,
        think_mode: ThinkMode = ThinkMode.not_think,
        response_format: BaseModel | None = None,
        **client_kwargs,
    ):
        """
        Retry a failed turn with feedback and traceback.
        """
        assert error_idx > 0, "error_idx must be greater than 0"
        prompt = RETRY_TEMPLATE.render(feedback=feedback, traceback=traceback)
        history = [t for t in self._history if t.id == turn_id]
        history_msg = []
        for turn in history:
            history_msg.extend(turn.message)
        response, message = await self.llm(
            prompt,
            history=history_msg,
            return_message=True,
            think_mode=think_mode,
            response_format=response_format,
            **client_kwargs,
        )
        turn = Turn(
            id=turn_id,
            prompt=prompt,
            response=response,
            message=message,
            retry=error_idx,
        )
        return await self.__post_process__(response, history, turn)

    async def __call__(
        self,
        think_mode: ThinkMode = ThinkMode.not_think,
        images: list[str] = None,
        recent: int = 0,
        response_format: BaseModel | None = None,
        client_kwargs: dict | None = None,
        **jinja_args,
    ):
        """
        Call the agent with prompt arguments.

        Args:
            images (list[str]): A list of image file paths.
            recent (int): The number of recent turns to include.
            similar (int): The number of similar turns to include.
            **jinja_args: Additional arguments for the Jinja2 template.

        Returns:
            The response from the role.
        """
        if isinstance(images, str):
            images = [images]
        assert self.prompt_args == set(jinja_args.keys()), (
            f"Invalid arguments, expected: {self.prompt_args}, got: {jinja_args.keys()}"
        )
        prompt = self.template.render(**jinja_args)
        history = await self.get_history(recent)
        history_msg = []
        for turn in history:
            history_msg.extend(turn.message)

        if client_kwargs is None:
            client_kwargs = {}
        response, message = await self.llm(
            prompt,
            think_mode=think_mode,
            system_message=self.system_message,
            history=history_msg,
            images=images,
            return_message=True,
            response_format=response_format,
            **client_kwargs,
        )
        turn = Turn(
            id=self.next_turn_id,
            prompt=prompt,
            response=response,
            message=message,
            images=images,
        )
        return turn.id, await self.__post_process__(response, history, turn)

    async def get_history(self, recent: int):
        """
        Get the conversation history.
        """
        history = self._history[-recent:] if recent > 0 else []
        history.sort(key=lambda x: x.id)
        return history

    async def __post_process__(
        self, response: str, history: list[Turn], turn: Turn
    ) -> str | dict:
        """
        Post-process the response from the agent.
        """
        self._history.append(turn)
        if self.record_cost:
            turn.calc_token()
            self.calc_cost(history + [turn])
        if self.return_json:
            response = get_json_from_response(response)
        return response


def calc_image_tokens(images: list[str]):
    """
    Calculate the number of tokens for a list of images.
    """
    tokens = 0
    for image in images:
        with open(image, "rb") as f:
            width, height = Image.open(f).size
        if width > 1024 or height > 1024:
            if width > height:
                height = int(height * 1024 / width)
                width = 1024
            else:
                width = int(width * 1024 / height)
                height = 1024
        h = ceil(height / 512)
        w = ceil(width / 512)
        tokens += 85 + 170 * h * w
    return tokens



from math import ceil
import os
from pathlib import Path
from random import shuffle
from pptagent.pptgen import PPTAgent
from pptagent.llms import AsyncLLM
from pptagent.presentation.layout import Layout
from pptagent.response.pptgen import (
    EditorOutput,
    SlideElement,
)
from pptagent.multimodal import ImageLabler
from pptagent.utils import Config, Language, package_join
from pptagent.presentation import Presentation
from glob import glob
from os.path import join, exists
import json
from fastmcp import FastMCP
from pptagent.utils import get_logger
from pptagent.pptgen import get_length_factor

logger = get_logger(__name__)


def mcp_slide_validate(editor_output: EditorOutput, layout: Layout, prs_lang: Language):
    warnings = []
    errors = []
    length_factor = get_length_factor(prs_lang, Language.english())
    layout_elements = {el.name for el in layout.elements}
    editor_elements = {el.name for el in editor_output.elements}
    for el in layout_elements - editor_elements:
        errors.append(f"Element {el} not found in editor output")
    for el in editor_elements - layout_elements:
        errors.append(f"Element {el} not found in layout")
    for el in layout.elements:
        if layout[el.name].type == "image":
            for i in range(len(editor_output[el.name].data)):
                if not exists(editor_output[el.name].data[i]):
                    errors.append(f"Image {editor_output[el.name].data[i]} not found")
        else:
            charater_counts = max([len(i) for i in editor_output[el.name].data])
            expected_length = ceil(layout[el.name].suggested_characters * length_factor)
            if charater_counts - expected_length > 5:
                warnings.append(
                    f"Element {el.name} has {charater_counts} characters, but the expected length is {expected_length}"
                )
    return warnings, errors


class PPTAgentServer(PPTAgent):
    roles = [
        "coder",
    ]

    def __init__(self):
        self.source_doc = None
        self.mcp = FastMCP("PPTAgent")
        self.slides = []
        self.layout: Layout | None = None
        self.editor_output: EditorOutput | None = None
        model = AsyncLLM(
            os.getenv("PPTAGENT_MODEL"),
            os.getenv("PPTAGENT_API_BASE"),
            os.getenv("PPTAGENT_API_KEY"),
        )
        if not model.to_sync().test_connection():
            msg = "Unable to connect to the model, please set the PPTAGENT_MODEL, PPTAGENT_API_BASE, and PPTAGENT_API_KEY environment variables correctly"
            logger.error(msg)
            raise Exception(msg)
        super().__init__(language_model=model, vision_model=model)
        # load templates, a directory containing pptx, json, and description for each template
        templates = glob(package_join("templates", "*/"))
        self.template_description = {}
        for template in templates:
            self.template_description[Path(template).name] = open(
                join(template, "description.txt")
            ).read()

        logger.info(
            f"{len(templates)} templates loaded:"
            + ", ".join(self.template_description.keys())
        )

    def register_tools(self):
        @self.mcp.tool()
        def list_available_templates() -> list[dict]:
            """List all available templates."""
            return {
                "message": "Available templates:",
                "templates": [
                    {
                        "name": template_name,
                        "description": self.template_description[template_name],
                    }
                    for template_name in self.template_description.keys()
                ],
            }

        @self.mcp.tool()
        def set_template(template_name: str = "default"):
            """Select a PowerPoint template by name.

            Args:
                template_name: The name of the template to select

            Returns:
                dict: Success message and list of available layouts
            """
            template_folder = package_join("templates", template_name)
            assert template_name in self.template_description, (
                f"Template {template_name} not available, please choose from {list(self.template_description.keys())}"
            )
            prs_config = Config(template_folder)
            prs = Presentation.from_file(
                join(template_folder, "source.pptx"), prs_config
            )
            image_labler = ImageLabler(prs, prs_config)
            image_labler.apply_stats(
                json.load(open(join(template_folder, "image_stats.json")))
            )
            self.set_reference(
                slide_induction=json.load(
                    open(join(template_folder, "slide_induction.json"))
                ),
                presentation=prs,
            )

            return {
                "message": "Template set successfully, please select layout from given layouts later",
                "template_description": self.template_description[template_name],
                "available_layouts": list(self.layouts.keys()),
            }

        @self.mcp.tool()
        async def set_layout(layout: str):
            """Select a layout for generating slides.

            Args:
                layout: Name of the layout to use. Must be one of the available layouts from set_template.

            Returns:
                dict: Success message, instructions, and content schema for the selected layout.
            """
            assert self._initialized, (
                "PPTAgent not initialized, please call `set_template` first"
            )
            assert layout in self.layouts, (
                "Given layout was not in available layouts: " + ", ".join(self.layouts)
            )
            if self.layout is not None:
                message = "Layout update from " + self.layout.title + " to " + layout
                message += "\nDid you forget to call `generate_slide` after setting slide content?"
            else:
                message = "Layout " + layout + " selected successfully"
            self.layout = self.layouts[layout]
            return {
                "message": message,
                "instructions": "Generate slide content strictly following the schema below",
                "schema": self.layout.content_schema,
            }

        @self.mcp.tool()
        async def set_slide_content(structured_slide_elements: list[dict]):
            """Set the slide elements for generating a PowerPoint slide.
            Note that this function will not generate a slide, you should call `generate_slide`.

            Args:
                structured_slide_elements: List of slide elements with their content
                should follow the content schema and adhere to
                [
                    {
                        "name": "element_name",
                        "data": ["content1", "content2", "..."]
                        // Array of strings for text elements
                        // OR array of image paths for image elements: ["/path/to/image1.jpg", "/path/to/image2.png"]
                    }
                ]
            Returns:
                dict: Success message, warnings, and errors
            """
            self.structured_slide_elements = structured_slide_elements
            assert self.layout is not None, (
                "Layout is not selected, please call `select_layout` before generating slide"
            )
            editor_output = EditorOutput(
                elements=[SlideElement(**e) for e in structured_slide_elements]
            )
            warnings, errors = mcp_slide_validate(
                editor_output, self.layout, self.reference_lang
            )
            if errors:
                raise ValueError("Errors:\n" + "\n".join(errors))

            self.editor_output = editor_output
            if warnings:
                return {
                    "message": "Slide elements set with warnings. Review warnings, consider resetting slide content, or proceed if acceptable.",
                    "warnings": warnings,
                }
            return {
                "message": "Slide elements set successfully. Ready to generate slide."
            }

        @self.mcp.tool()
        async def generate_slide():
            """Generate a PowerPoint slide after layout and slide elements are set.

            Returns:
                dict: Success message with slide number and next steps
            """
            if self.editor_output is None:
                raise ValueError(
                    "Slide elements are not set, please call `set_slide_content` before generating slide"
                )

            command_list, template_id = self._generate_commands(
                self.editor_output, self.layout
            )
            slide, _ = await self._edit_slide(command_list, template_id)

            # Reset state after successful generation
            self.layout = None
            self.editor_output = None
            self.slides.append(slide)

            slide_number = len(self.slides)
            available_layouts = list(self.layouts.keys())
            shuffle(available_layouts)

            return {
                "message": f"Slide {slide_number:02d} generated successfully",
                "next_steps": "You can now save the slides or continue generating more slides",
                "available_layouts": available_layouts,
            }

        @self.mcp.tool()
        async def save_generated_slides(pptx_path: str):
            """Save the generated slides to a PowerPoint file.

            Args:
                pptx_path: The path to save the PowerPoint file
            """
            assert len(self.slides), (
                "No slides generated, please call `generate_slide` first"
            )
            os.makedirs(os.path.dirname(pptx_path), exist_ok=True)
            self.empty_prs.slides = self.slides
            self.empty_prs.save(pptx_path)
            self.slides = []
            self._initialized = False
            return f"total {len(self.empty_prs.slides)} slides saved to {pptx_path}"


def main():
    server = PPTAgentServer()
    server.register_tools()
    server.mcp.run(show_banner=False)


if __name__ == "__main__":
    main()



import hashlib
import re

from jinja2 import Environment, StrictUndefined
from PIL import Image
from pydantic import BaseModel, Field, create_model

from pptagent.llms import AsyncLLM
from os.path import exists, join
from pptagent.utils import (
    edit_distance,
    get_html_table_image,
    get_logger,
    package_join,
)

from .doc_utils import parse_table_with_merges

env = Environment(undefined=StrictUndefined)

IMAGE_PARSING_REGEX = re.compile(r"\((.*?)\)")
TABLE_CAPTION_PROMPT = env.from_string(
    open(
        package_join("prompts", "document", "markdown_table_caption.txt"),
        encoding="utf-8",
    ).read()
)
IMAGE_CAPTION_PROMPT = env.from_string(
    open(
        package_join("prompts", "document", "markdown_image_caption.txt"),
        encoding="utf-8",
    ).read()
)

logger = get_logger(__name__)


class Media(BaseModel):
    markdown_content: str
    near_chunks: tuple[str, str]
    path: str | None = None
    caption: str | None = None

    @property
    def size(self):
        assert self.path is not None, "Path is required to get size"
        return Image.open(self.path).size

    def parse(self, image_dir: str):
        """
        Parse the markdown content to extract image path and alt text.
        Format expected: ![alt text](image.png)
        """
        match = IMAGE_PARSING_REGEX.search(self.markdown_content)
        if match is None:
            raise ValueError("No image found in the markdown content")
        image_path = match.group(1)
        if not exists(image_path):
            image_path = join(image_dir, image_path)
        assert exists(image_path), f"image file not found: {image_path}"
        self.path = image_path

    async def get_caption(self, vision_model: AsyncLLM):
        assert self.path is not None, "Path is required to get caption"
        if self.caption is None:
            self.caption = await vision_model(
                IMAGE_CAPTION_PROMPT.render(
                    markdown_caption=self.near_chunks,
                ),
                self.path,
            )
            logger.debug(f"Caption: {self.caption}")


class Table(Media):
    cells: list[list[str]] | None = None
    merge_area: list[tuple[int, int, int, int]] | None = None

    def parse(self, image_dir: str):
        cells, merges = parse_table_with_merges(self.markdown_content)
        self.cells = cells
        self.merge_area = merges

        if self.path is None:
            self.path = join(
                image_dir,
                f"table_{hashlib.md5(str(self.cells).encode()).hexdigest()[:4]}.png",
            )
        get_html_table_image(self.markdown_content, self.path)

    async def get_caption(self, language_model: AsyncLLM):
        if self.caption is None:
            self.caption = await language_model(
                TABLE_CAPTION_PROMPT.render(
                    markdown_content=self.markdown_content,
                    markdown_caption=self.near_chunks,
                )
            )
            logger.debug(f"Caption: {self.caption}")


class SubSection(BaseModel):
    title: str
    content: str


class Metadata(BaseModel):
    name: str
    value: str


class Section(BaseModel):
    title: str
    summary: str
    content: list[SubSection | Media | Table]
    markdown_content: str | None = None

    def iter_medias(self):
        for block in self.content:
            if isinstance(block, Media):
                yield block

    @classmethod
    def response_model(cls):
        # Create a new model with only SubSection blocks and metadata
        return create_model(
            cls.__name__,
            title=(str, Field(...)),
            summary=(str, Field(...)),
            subsections=(list[SubSection], Field(...)),
            metadata=(list[Metadata], Field(default_factory=list)),
            __base__=BaseModel,
        )

    def __getitem__(self, key: str):
        for block in self.content:
            if isinstance(block, SubSection) and block.title == key:
                return block
            elif isinstance(block, Media) and (
                block.path == key or block.caption == key
            ):
                return block
        raise KeyError(f"No subsection or media with title {key} found")


def link_medias(
    medias: list[dict],
    section: Section,
    max_chunk_size: int = 256,
):
    """
    Link media elements to the section by inserting them into the blocks list at appropriate positions.

    Args:
        medias: List of media dictionaries (tables, images)
        section: Section object to insert medias into
        max_chunk_size: Maximum size of text chunk to consider for matching
    """
    if not medias:
        return

    # Convert media dictionaries to Media instances
    media_instances = []
    for media_dict in medias:
        if media_dict.get("type") == "table":
            media_instances.append(Table(**media_dict))
        else:
            media_instances.append(Media(**media_dict))

    # Find the best insertion position for each media
    for media in media_instances:
        if len(media.near_chunks[0]) < max_chunk_size:
            # If context is small, insert at the beginning
            section.content.insert(0, media)
        else:
            # Find the most similar SubSection based on content
            best_match_idx = 0
            best_similarity = 0

            for i, block in enumerate(section.content):
                if isinstance(block, SubSection):
                    similarity = edit_distance(media.near_chunks[0], block.content)
                    if similarity > best_similarity:
                        best_similarity = similarity
                        best_match_idx = i

            section.content.insert(best_match_idx + 1, media)



from test.conftest import test_config

from bs4 import BeautifulSoup
from pptx import Presentation

from pptagent.apis import (
    API_TYPES,
    CodeExecutor,
    markdown,
    process_element,
    replace_para,
)


def test_api_docs():
    executor = CodeExecutor(3)
    docs = executor.get_apis_docs(API_TYPES.Agent.value)
    assert len(docs) > 0


def test_replace_para():
    text = "这是一个**加粗和*斜体*文本**，还有*斜体和`Code def a+b`*，~~删除~~，[链接](http://example.com)"
    prs = Presentation(test_config.ppt)
    slide = prs.slides[0]
    replace_para(0, text, slide.shapes[0])
    runs = slide.shapes[0].text_frame.paragraphs[0].runs
    assert runs[1].font.bold
    assert runs[2].font.bold and runs[2].font.italic
    assert runs[6].font.name == "Consolas"
    assert runs[8].font.strikethrough
    assert runs[10].hyperlink.address == "http://example.com"


def test_list_parsing():
    text = """
    - 项目1
    - 项目2

    1. 项目1
    2. 项目2
    """
    html = markdown(text).strip()
    soup = BeautifulSoup(html, "html.parser")
    blocks = process_element(soup)
    assert len(blocks) == 1
    assert "ol" not in html and "ul" not in html



import asyncio

import PIL.Image

from pptagent.llms import LLM, AsyncLLM
from pptagent.presentation import Picture, Presentation
from os.path import basename, join
from pptagent.utils import Config, get_logger, package_join

logger = get_logger(__name__)


class ImageLabler:
    """
    A class to extract images information, including caption, size, and appearance times in a presentation.
    """

    def __init__(self, presentation: Presentation, config: Config):
        """
        Initialize the ImageLabler.

        Args:
            presentation (Presentation): The presentation object.
            config (Config): The configuration object.
        """
        self.presentation = presentation
        self.slide_area = presentation.slide_width.pt * presentation.slide_height.pt
        self.image_stats = {}
        self.config = config
        self.collect_images()

    def apply_stats(self, image_stats: dict[str, dict] | None = None):
        """
        Apply image captions to the presentation.
        """
        if image_stats is None:
            image_stats = self.image_stats

        for slide in self.presentation.slides:
            for shape in slide.shape_filter(Picture):
                if shape.caption is None:
                    caption = image_stats[basename(shape.img_path)]["caption"]
                    shape.caption = max(caption.split("\n"), key=len)

    async def caption_images_async(self, vision_model: AsyncLLM):
        """
        Generate captions for images in the presentation asynchronously.

        Args:
            vision_model (AsyncLLM): The async vision model to use for captioning.

        Returns:
            dict: Dictionary containing image stats with captions.
        """
        assert isinstance(vision_model, AsyncLLM), (
            "vision_model must be an AsyncLLM instance"
        )
        caption_prompt = open(
            package_join("prompts", "caption.txt"), encoding="utf-8"
        ).read()

        async with asyncio.TaskGroup() as tg:
            for image, stats in self.image_stats.items():
                if "caption" not in stats:
                    task = tg.create_task(
                        vision_model(
                            caption_prompt,
                            join(self.config.IMAGE_DIR, image),
                        )
                    )
                    task.add_done_callback(
                        lambda t, image=image: (
                            self.image_stats[image].update({"caption": t.result()}),
                            logger.debug("captioned %s: %s", image, t.result()),
                        )
                    )

        self.apply_stats()
        return self.image_stats

    def caption_images(self, vision_model: LLM):
        """
        Generate captions for images in the presentation.

        Args:
            vision_model (LLM): The vision model to use for captioning.

        Returns:
            dict: Dictionary containing image stats with captions.
        """
        assert isinstance(vision_model, LLM), "vision_model must be an LLM instance"
        caption_prompt = open(
            package_join("prompts", "caption.txt"), encoding="utf-8"
        ).read()
        for image, stats in self.image_stats.items():
            if "caption" not in stats:
                stats["caption"] = vision_model(
                    caption_prompt, join(self.config.IMAGE_DIR, image)
                )
                logger.debug("captioned %s: %s", image, stats["caption"])
        self.apply_stats()
        return self.image_stats

    def collect_images(self):
        """
        Collect images from the presentation and gather other information.
        """
        for slide_index, slide in enumerate(self.presentation.slides):
            for shape in slide.shape_filter(Picture):
                image_path = basename(shape.img_path)
                if image_path == "pic_placeholder.png":
                    continue
                if image_path not in self.image_stats:
                    size = PIL.Image.open(join(self.config.IMAGE_DIR, image_path)).size
                    self.image_stats[image_path] = {
                        "size": size,
                        "appear_times": 0,
                        "slide_numbers": set(),
                        "relative_area": shape.area / self.slide_area * 100,
                    }
                self.image_stats[image_path]["appear_times"] += 1
                self.image_stats[image_path]["slide_numbers"].add(slide_index + 1)
        for image_path, stats in self.image_stats.items():
            stats["slide_numbers"] = sorted(list(stats["slide_numbers"]))
            ranges = self._find_ranges(stats["slide_numbers"])
            top_ranges = sorted(ranges, key=lambda x: x[1] - x[0], reverse=True)[:3]
            top_ranges_str = ", ".join(
                [f"{r[0]}-{r[1]}" if r[0] != r[1] else f"{r[0]}" for r in top_ranges]
            )
            stats["top_ranges_str"] = top_ranges_str

    def _find_ranges(self, numbers):
        """
        Find consecutive ranges in a list of numbers.
        """
        ranges = []
        start = numbers[0]
        end = numbers[0]
        for num in numbers[1:]:
            if num == end + 1:
                end = num
            else:
                ranges.append((start, end))
                start = num
                end = num
        ranges.append((start, end))
        return ranges



from .doc_utils import get_tree_structure
from .document import Document
from .element import Media, Section, SubSection, Table

__all__ = [
    "Document",
    "Media",
    "Section",
    "SubSection",
    "Table",
    "get_tree_structure",
]



from test.conftest import test_config

import pytest

from pptagent.document import Document
from pptagent.multimodal import ImageLabler
from pptagent.pptgen import PPTAgent
from pptagent.presentation import Presentation
from os.path import join


@pytest.mark.asyncio
@pytest.mark.llm
async def test_pptgen():
    pptgen = PPTAgent(
        language_model=test_config.language_model,
        vision_model=test_config.vision_model,
    ).set_reference(
        presentation=Presentation.from_file(
            join(test_config.template, "source.pptx"), test_config.config
        ),
        slide_induction=test_config.get_slide_induction(),
    )
    labeler = ImageLabler(pptgen.presentation, test_config.config)
    labeler.apply_stats(test_config.get_image_stats())

    document = Document(**test_config.get_document_json())
    await pptgen.generate_pres(document, 3)



from copy import deepcopy
from test.conftest import test_config

import pytest


@pytest.mark.asyncio
@pytest.mark.llm
async def test_batchllm():
    """
    Test the async LLM connection and functionality.
    """
    # Test connection
    test_config.language_model.use_batch = True
    result = await deepcopy(test_config.language_model).test_connection()
    assert result is True, "Async LLM connection failed"

    # Test response generation
    response = await test_config.language_model("Hello, how are you?", max_tokens=1)
    assert response is not None, "Async LLM returned None response"
    assert len(response) > 0, "Async LLM returned empty response"


@pytest.mark.asyncio
@pytest.mark.llm
async def test_asyncllm():
    """
    Test the async LLM connection and functionality.
    """
    # Test connection
    result = await test_config.language_model.test_connection()
    assert result is True, "Async LLM connection failed"

    # Test response generation
    response = await test_config.language_model("Hello, how are you?", max_tokens=1)
    assert response is not None, "Async LLM returned None response"
    assert len(response) > 0, "Async LLM returned empty response"


@pytest.mark.llm
def test_sync_llm():
    """
    Test the synchronous LLM connection and functionality.
    """

    sync_language_model = test_config.language_model.to_sync()
    result = sync_language_model.test_connection()
    assert result is True, "Sync LLM connection failed"
    response = sync_language_model("Hello, how are you?", max_tokens=1)
    assert response is not None, "Sync LLM returned None response"
    assert len(response) > 0, "Sync LLM returned empty response"



import tempfile
import traceback
from collections.abc import Generator
from dataclasses import dataclass
from functools import partial
from typing import Literal

from pptx import Presentation as load_prs
from pptx.enum.shapes import MSO_SHAPE_TYPE
from pptx.shapes.base import BaseShape
from pptx.shapes.group import GroupShape as PPTXGroupShape
from pptx.slide import Slide as PPTXSlide

from pptagent.utils import Config, get_logger, package_join

from .shapes import (
    Background,
    Closure,
    ClosureType,
    GroupShape,
    Paragraph,
    Picture,
    ShapeElement,
    StyleArg,
)

# Type variable for ShapeElement subclasses

logger = get_logger(__name__)


@dataclass
class SlidePage:
    """
    A class to represent a slide page in a presentation.
    """

    shapes: list[ShapeElement]
    backgrounds: list[Background]
    slide_idx: int
    real_idx: int
    slide_notes: str | None
    slide_layout_name: str | None
    slide_title: str | None
    slide_width: int
    slide_height: int

    def __post_init__(self):
        # Assign group labels to group shapes
        groups_shapes_labels = []
        for shape in self.shape_filter(GroupShape):
            for group_shape in groups_shapes_labels:
                if group_shape == shape:
                    shape.group_label = group_shape.group_label
                    continue
            groups_shapes_labels.append(shape)
            shape.group_label = f"group_{len(groups_shapes_labels)}"

    @classmethod
    def from_slide(
        cls,
        slide: PPTXSlide,
        slide_idx: int,
        real_idx: int,
        slide_width: int,
        slide_height: int,
        config: Config,
        shape_cast: dict[MSO_SHAPE_TYPE, type[ShapeElement] | None],
    ) -> "SlidePage":
        """
        Create a SlidePage from a PPTXSlide.

        Args:
            slide (PPTXSlide): The slide object.
            slide_idx (int): The index of the slide.
            real_idx (int): The real index of the slide.
            slide_width (int): The width of the slide.
            slide_height (int): The height of the slide.
            config (Config): The configuration object.
            shape_cast (dict[MSO_SHAPE_TYPE, type[ShapeElement] | None]): Mapping of shape types to their corresponding ShapeElement classes.
            Set the value to None for any MSO_SHAPE_TYPE to exclude that shape type from processing.
        Returns:
            SlidePage: The created SlidePage.
        """
        backgrounds = [Background.from_slide(slide, config)]
        shapes = []
        for i, shape in enumerate(slide.shapes):
            if not shape.visible:
                continue
            if shape_cast.get(shape.shape_type, -1) is None:
                continue
            shapes.append(
                ShapeElement.from_shape(
                    slide_idx, i, shape, config, slide_width * slide_height, shape_cast
                )
            )
        for i, s in enumerate(shapes):
            if isinstance(s, Picture) and s.area / s.slide_area > 0.95:
                backgrounds.append(shapes.pop(i))

        slide_layout_name = slide.slide_layout.name if slide.slide_layout else None
        slide_title = slide.shapes.title.text if slide.shapes.title else None
        slide_notes = (
            slide.notes_slide.notes_text_frame.text
            if slide.has_notes_slide and slide.notes_slide.notes_text_frame
            else None
        )

        return cls(
            shapes,
            backgrounds,
            slide_idx,
            real_idx,
            slide_notes,
            slide_layout_name,
            slide_title,
            slide_width,
            slide_height,
        )

    def build(self, slide: PPTXSlide) -> PPTXSlide:
        """
        Build the slide page in a slide.

        Args:
            slide (PPTXSlide): The slide to build the slide page in.

        Returns:
            PPTXSlide: The built slide.
        """
        # Remove existing placeholders
        for ph in slide.placeholders:
            ph.element.getparent().remove(ph.element)

        # Build backgrounds, shapes and apply closures
        for shape in sorted(self.backgrounds + list(self), key=lambda x: x.shape_idx):
            build_shape = shape.build(slide)
            for closure in shape.closures:
                try:
                    closure.apply(build_shape)
                except Exception as e:
                    raise ValueError(f"Failed to apply closures to slides: {e}")
        return slide

    def iter_paragraphs(self) -> Generator[Paragraph, None, None]:
        for shape in self:  # this considered the group shapes
            if not shape.text_frame.is_textframe:
                continue
            for para in shape.text_frame.paragraphs:
                if para.idx != -1:
                    yield para

    def shape_filter(
        self,
        shape_type: type[ShapeElement],
        from_groupshape: bool = True,
        return_father: bool = False,
    ) -> (
        Generator[ShapeElement, None, None]
        | Generator[tuple["SlidePage", ShapeElement], None, None]
    ):
        """
        Filter shapes in the slide by type.

        Args:
            shape_type (Type[ShapeElement]): The type of shapes to filter.
            shapes (Optional[List[ShapeElement]]): The shapes to filter.

        Yields:
            ShapeElement: The filtered shapes.
        """
        for shape in self.shapes:
            if isinstance(shape, shape_type):
                if return_father:
                    yield (self, shape)
                else:
                    yield shape
            elif from_groupshape and isinstance(shape, GroupShape):
                yield from shape.shape_filter(shape_type, return_father)

    def get_content_type(self) -> Literal["text", "image"]:
        """
        Get the content type of the slide.

        Returns:
            Literal["text", "image"]: The content type of the slide.
        """
        if len(list(self.shape_filter(Picture))) == 0:
            return "text"
        return "image"

    def to_html(self, style_args: StyleArg | None = None, **kwargs) -> str:
        """
        Represent the slide page in HTML.

        Args:
            style_args (Optional[StyleArg]): The style arguments for HTML conversion.
            **kwargs: Additional arguments.

        Returns:
            str: The HTML representation of the slide page.
        """
        if style_args is None:
            style_args = StyleArg(**kwargs)
        shapes_html = [shape.to_html(style_args) for shape in self.shapes]
        shapes_html = [html for html in shapes_html if html]
        return "".join(
            [
                "<!DOCTYPE html>\n<html>\n",
                (f"<title>{self.slide_title}</title>\n" if self.slide_title else ""),
                f'<body style="width:{self.slide_width}pt; height:{self.slide_height}pt;">\n',
                "\n".join(shapes_html),
                "</body>\n</html>\n",
            ]
        )

    def to_text(self, show_image: bool = False) -> str:
        """
        Represent the slide page in text.

        Args:
            show_image (bool): Whether to show image captions.

        Returns:
            str: The text representation of the slide page.

        Raises:
            ValueError: If an image caption is not found.
        """
        text_content = ""
        for para in self.iter_paragraphs():
            if not para.text:
                continue
            if para.bullet:
                text_content += para.bullet
            text_content += para.text + "\n"
        if show_image:
            for image in self.shape_filter(Picture):
                text_content += "\n" + "Image: " + image.caption
        return text_content

    def __iter__(self):
        """
        Iterate over all shapes in the slide page.

        Yields:
            ShapeElement: Each shape in the slide page.
        """
        for shape in self.shapes:
            if isinstance(shape, GroupShape):
                yield from shape
            else:
                yield shape

    def __len__(self) -> int:
        """
        Get the number of shapes in the slide page.

        Returns:
            int: The number of shapes.
        """
        return len(self.shapes)


@dataclass
class Presentation:
    """
    PPTAgent's representation of a presentation.
    Aiming at a more readable and editable interface.
    """

    slides: list[SlidePage]
    error_history: list[tuple[int, str]]
    slide_width: float
    slide_height: float
    source_file: str
    num_pages: int

    def __post_init__(self):
        self.prs = load_prs(self.source_file)
        self.layout_mapping = {layout.name: layout for layout in self.prs.slide_layouts}
        self.prs.core_properties.last_modified_by = "PPTAgent"

    @classmethod
    def from_file(
        cls,
        file_path: str,
        config: Config | None = None,
        shape_cast: dict[MSO_SHAPE_TYPE, type[ShapeElement]] | None = None,
    ) -> "Presentation":
        """
        Parse a Presentation from a file.

        Args:
            file_path (str): The path to the presentation file.
            config (Config): The configuration object.
            shape_cast (dict[MSO_SHAPE_TYPE, type[ShapeElement]] | None): Optional mapping of shape types to their corresponding ShapeElement classes.
            Set the value to None for any MSO_SHAPE_TYPE to exclude that shape type from processing.
        Returns:
            Presentation: The parsed Presentation.
        """
        if config is None:
            config = Config(tempfile.mkdtemp())
        prs = load_prs(file_path)
        slide_width = prs.slide_width
        slide_height = prs.slide_height
        slides = []
        error_history = []
        slide_idx = 0
        layouts = [layout.name for layout in prs.slide_layouts]
        num_pages = len(prs.slides)

        if shape_cast is None:
            shape_cast = {}

        for slide in prs.slides:
            # Skip slides that won't be printed to PDF, as they are invisible
            if slide._element.get("show", 1) == "0":
                continue

            slide_idx += 1
            try:
                if slide.slide_layout.name not in layouts:
                    raise ValueError(
                        f"Slide layout {slide.slide_layout.name} not found"
                    )
                slides.append(
                    SlidePage.from_slide(
                        slide,
                        slide_idx - len(error_history),
                        slide_idx,
                        slide_width.pt,
                        slide_height.pt,
                        config,
                        shape_cast,
                    )
                )
            except Exception as e:
                error_history.append((slide_idx, str(e)))
                logger.error(
                    "Fail to parse slide %d of %s: %s",
                    slide_idx,
                    file_path,
                    e,
                )
                logger.error(traceback.format_exc())

        return cls(
            slides, error_history, slide_width, slide_height, file_path, num_pages
        )

    def save(self, file_path: str, layout_only: bool = False) -> None:
        """
        Save the presentation to a file.

        Args:
            file_path (str): The path to save the presentation to.
            layout_only (bool): Whether to save only the layout.
        """
        self.clear_slides()
        for slide in self.slides:
            if layout_only:
                self.clear_images(slide.shapes)
            pptx_slide = self.build_slide(slide)
            if layout_only:
                self.clear_text(pptx_slide.shapes)
        self.prs.save(file_path)

    def build_slide(self, slide: SlidePage) -> PPTXSlide:
        """
        Build a slide in the presentation.
        """
        return slide.build(
            self.prs.slides.add_slide(self.layout_mapping[slide.slide_layout_name])
        )

    def validate(self, slide: SlidePage) -> PPTXSlide:
        """
        Build a slide in the presentation.
        """

        from pptagent.apis import del_para

        for shape in slide:
            if not shape.text_frame.is_textframe:
                continue
            for para in shape.text_frame.paragraphs:
                if not para.edited and para.idx != -1:
                    shape._closures[ClosureType.POST_PROCESS].append(
                        Closure(
                            partial(del_para, para.real_idx),
                            para.real_idx,
                        )
                    )

        return self.build_slide(slide)

    def clear_slides(self):
        """
        Delete all slides from the presentation.
        """
        while len(self.prs.slides) != 0:
            rId = self.prs.slides._sldIdLst[0].rId
            self.prs.part.drop_rel(rId)
            del self.prs.slides._sldIdLst[0]

    def clear_images(self, shapes: list[ShapeElement]):
        for shape in shapes:
            if isinstance(shape, GroupShape):
                self.clear_images(shape.shapes)
            elif isinstance(shape, Picture):
                shape.img_path = package_join("resource", "pic_placeholder.png")

    def clear_text(self, shapes: list[BaseShape]):
        for shape in shapes:
            if isinstance(shape, PPTXGroupShape):
                self.clear_text(shape.shapes)
            elif shape.has_text_frame:
                for para in shape.text_frame.paragraphs:
                    for run in para.runs:
                        run.text = "a" * len(run.text)

    def to_text(self, show_image: bool = False) -> str:
        """
        Represent the presentation in text.
        """
        return "\n----\n".join(
            [
                (
                    f"Slide {slide.slide_idx} of {len(self.slides)}\n"
                    + (f"Title:{slide.slide_title}\n" if slide.slide_title else "")
                    + slide.to_text(show_image)
                )
                for slide in self.slides
            ]
        )

    def __iter__(self):
        yield from self.slides

    def __len__(self) -> int:
        """
        Get the number of slides in the presentation.
        """
        return len(self.slides)

    def __getstate__(self) -> object:
        state = self.__dict__.copy()
        state["prs"] = None
        state["layout_mapping"] = None
        return state

    def __setstate__(self, state: object):
        self.__dict__.update(state)
        self.prs = load_prs(self.source_file)
        self.layout_mapping = {layout.name: layout for layout in self.prs.slide_layouts}
